{-# LANGUAGE TypeOperators, 
             FlexibleInstances,
             MultiParamTypeClasses, 
             FlexibleContexts, 
             UndecidableInstances,
             OverlappingInstances,
             DeriveFunctor,
             GADTs,
             TemplateHaskell,
             StandaloneDeriving,
             KindSignatures #-}

-----------------------------------
-- Evaluator for Source Language --
-----------------------------------


module Evaluator where



import Control.Monad
import Control.Monad.State hiding (State)
import qualified Data.Map as Map

import HFixpoint
import SourceLanguage


getRef :: (MonadState St m, MonadPlus m) => Ref -> m Int
getRef v = do s <- get
              case Map.lookup v s of
                Just n -> return n
                Nothing -> mzero


setRef :: MonadState St m => Ref -> Int -> m ()
setRef v n = modify (Map.insert v n)


data VNum (e :: * -> *) l where
    Num   :: Int -> VNum e IExp
data VBool (e :: * -> *) l where
    Bool  :: Bool -> VBool e BExp
data VUnit (e :: * -> *) l where
    Unit  :: VUnit e Stmt

num :: (VNum ::< f) => Int -> HFix f IExp
num = hinject . Num
bool :: (VBool ::< f) => Bool -> HFix f BExp
bool = hinject . Bool
unit :: (VUnit ::< f) => HFix f Stmt
unit = hinject Unit

getNum :: (VNum ::< f) => HFix f IExp -> Maybe Int
getNum (HIn e) = case hprj e of 
                   Just (Num n) -> Just n
                   _ -> Nothing

getBool :: (VBool ::< f) => HFix f BExp -> Maybe Bool
getBool (HIn e) = case hprj e of 
                    Just (Bool b) -> Just b
                    _ -> Nothing

instance Show (VUnit i l) where
    show Unit = "Unit"
instance Show (VNum i l) where
    show (Num i) = show i
instance Show (VBool i l) where
    show (Bool i) = show i
  
class (HFunctor f, Monad m) => Eval f m v where
  evAlg :: f (m :. HFix v) i -> m (HFix v i)

instance (Eval f m v, Eval g m v) => Eval (f ::+ g) m v where
  evAlg (HInl x) = evAlg x
  evAlg (HInr y) = evAlg y

instance (Monad m, VNum ::< v) => Eval Arith m v where
  evAlg (Val n)   = return (num n)  
  evAlg (Add (C x) (C y)) = do Just n <- liftM getNum x
                               Just m <- liftM getNum y
                               return (num (n + m))
  evAlg (Mul (C x) (C y)) = do Just n <- liftM getNum x
                               Just m <- liftM getNum y
                               return (num (n * m))
instance MonadPlus m => Eval Except m v where
  evAlg (Throw)     = mzero
  evAlg (Catch (C x) (C h)) = x `mplus` h

instance (Monad m, VNum ::< v, VBool ::< v) => Eval Comp m v where
    evAlg (Equ (C xm) (C ym)) = do Just x <- liftM getNum xm
                                   Just y <- liftM getNum ym
                                   return (bool (x == y))
    evAlg (Lt (C xm) (C ym)) = do Just x <- liftM getNum xm
                                  Just y <- liftM getNum ym
                                  return (bool (x < y))
                                  
instance (Monad m, VBool ::< v) => Eval Boolean m v where
    evAlg (BVal b) = return (bool b)
    evAlg (And (C xm) (C ym)) = do Just x <- liftM getBool xm
                                   Just y <- liftM getBool ym
                                   return (bool (x && y))
    evAlg (Or (C xm) (C ym)) = do Just x <- liftM getBool xm
                                  Just y <- liftM getBool ym
                                  return (bool (x || y))


instance (MonadPlus m , MonadState St m, VNum ::< v, VUnit ::< v) => Eval State m v where
  evAlg (Get v)           = liftM num  (getRef v)
  evAlg (Set v (C m)) = liftM getNum m >>= \ (Just n) ->
                        setRef v n >> return unit
          
instance (Monad m, VBool ::< v, VUnit ::< v) => Eval While m v where
  evAlg (While (C c) (C lp)) = let loop = do Just b <- liftM getBool c
                                             if b then (lp >> loop)
                                                  else return unit
                               in loop

instance (Monad m, VBool ::< v) => Eval If m v where
  evAlg (If (C c) (C x) (C y)) = do Just b <- liftM getBool c
                                    if b then x else y
                    
instance Monad m => Eval Seq m v where
  evAlg (Seq (C prima) (C secunda)) = prima >> secunda

type Value = VNum ::+ VBool ::+ VUnit

eval' :: Eval f m v => HFix f :-> m :. HFix v
eval' = hfold (C . evAlg)

eval :: Eval f m v => HFix f i -> m (HFix v i)
eval = unC . eval'
