{-# LANGUAGE TypeOperators #-}


-----------------------------
-- Example source programs --
-----------------------------

module Eamples where


import Control.Monad.State hiding (State)
import qualified Data.Map as Map
import Prelude hiding (rem)


import Evaluator
import Compiler
import HFixpoint
import Fixpoint
import Graph
import TargetLanguage
import SourceLanguage

------------------------
-- Example: Degenerate -
------------------------

degen :: HFix (While ::+ Arith ::+ Except ::+ Boolean) Stmt
degen = whileF (bvalF False) throwF

compileDegen :: Graph (ARITH :+: COND :+: HALT :+: EXCEPT :+: BOOL)
compileDegen = comp degen

-- [v1 -> PUSHB False; JPC (THROW); HALT]

degen' :: HFix (While ::+ Arith ::+ State ::+ Boolean) Stmt
degen' = whileF (bvalF False) (setF x (valF 1337))


compileDegen' :: Graph (ARITH :+: COND :+: HALT :+: STATE :+: BOOL)
compileDegen' = comp degen'

-- [v1 -> PUSHB False; JPC (PUSH 1337; SET x; v1); HALT]


------------------------
-- Example: Factorial --
------------------------             


fac :: HFix (Seq ::+ Arith ::+ While ::+ State ::+ Comp) Stmt
fac =  setF y (valF 1) `seqF`
         whileF (valF 0 `ltF` getF x)
              ( seqF
                  (setF y (mulF (getF y) (getF x)   ) )
                  (setF x (addF (getF x) (valF (-1))) )
              )

evalFac :: Int -> Maybe (HFix Value Stmt, St)
evalFac n = runStateT res (Map.fromList [(x,n)])
    where res :: StateT St Maybe (HFix Value Stmt)
          res = eval fac

compileFac :: Graph (ARITH :+: COND :+: STATE :+: HALT :+: COMP)
compileFac = comp fac

-- PUSH 1; SET y; [v1 -> PUSH 0; GET x; LT; JPC (GET y; GET x; MUL;
-- SET y; GET x; PUSH (-1); ADD; SET x; v1); HALT]

evalFac' :: Int -> Maybe (HFix Value IExp, St)
evalFac' n = runStateT res (Map.fromList [(x,n)])
    where res :: StateT St Maybe (HFix Value IExp)
          res = eval fac' 

fac' :: HFix (Seq ::+ Arith ::+ While ::+ State ::+ Comp) IExp
fac' =  setF y (valF 1) `seqF`
          whileF (valF 0 `ltF` getF x)
              ( seqF
                  (setF y (mulF (getF y) (getF x)   ) )
                  (setF x (addF (getF x) (valF (-1))) )
              )
         `seqF` getF y


compileFac' :: Graph (ARITH :+: COND :+: STATE :+: HALT :+: COMP)
compileFac' = comp fac'

--- PUSH 1; SET y; [v1 -> PUSH 0; GET x; LT; JPC (GET y; GET x; MUL;
--- SET y; GET x; PUSH (-1); ADD; SET x; v1); GET y; HALT]


------------------------
-- Example: Remainder --
------------------------             

-- computes the remainder of the division x / y, returning -1 if y is
-- not positive

rem :: HFix (Seq ::+ Arith ::+ While ::+ State ::+ Comp ::+ If ::+ Boolean)  Stmt
rem =  ifF (getF y `ltF` valF 1) -- return -1 if y is not positive
            (setF x (valF (-1)))
            (setF z (getF y `mulF` valF (-1)) `seqF`
             whileF ((getF y `ltF` getF x) `orF` (getF y `equF` getF x))
                    (setF x (getF x `addF` getF z)))

evalRem :: Int -> Int -> Maybe (HFix Value Stmt, St)
evalRem n m = runStateT res (Map.fromList [(x,n), (y,m)])
    where res :: StateT St Maybe (HFix Value Stmt)
          res = eval rem

compileRem :: Graph (ARITH :+: COND :+: STATE :+: HALT :+: COMP :+: BOOL)
compileRem = comp rem

-- [v1 -> GET y; PUSH 1; LT; JPC (PUSH (-1); SET x; v2); GET y; PUSH
-- (-1); MUL; SET z; [v3 -> GET y; GET x; LT; GET y; GET x; EQ; OR;
-- JPC (GET x; GET z; ADD; SET x; v3); v2],v2 -> HALT]


-- computes the remainder of the division x / y, throwing an exception
-- if y is not positive

rem' :: HFix (Seq ::+ Arith ::+ While ::+ State ::+ Comp ::+ If ::+ Except ::+ Boolean)  Stmt
rem' =  ifF (getF y `ltF` valF 1) -- throw exception if y is not positive
            throwF
            (setF z (getF y `mulF` valF (-1)) `seqF`
             whileF ((getF y `ltF` getF x) `orF` (getF y `equF` getF x))
                    (setF x (getF x `addF` getF z)))

evalRem' :: Int -> Int -> Maybe (HFix Value Stmt, St)
evalRem' n m = runStateT res (Map.fromList [(x,n), (y,m)])
    where res :: StateT St Maybe (HFix Value Stmt)
          res = eval rem'

compileRem' :: Graph (ARITH :+: COND :+: STATE :+: HALT :+: COMP :+: BOOL :+: EXCEPT)
compileRem' = comp rem'

-- [v1 -> GET y; PUSH 1; LT; JPC (THROW); GET y; PUSH (-1); MUL; SET
-- z; [v3 -> GET y; GET x; LT; GET y; GET x; EQ; OR; JPC (GET x; GET
-- z; ADD; SET x; v3); v2],v2 -> HALT]
