{-# LANGUAGE Rank2Types, TypeOperators, FlexibleContexts, UndecidableInstances #-}

module Graph where

import Fixpoint

-----------------------
-- Structured Graphs --
-----------------------
 
data GraphT f v = Var v
                | Mu ([v] -> [GraphT f v])
                | GIn (f (GraphT f v))
newtype Graph f = Graph {unGraph :: forall v . GraphT f v}

letx     :: GraphT f v -> (v -> GraphT f v) -> GraphT f v
letx g f = Mu (\ ~(_ : x : _) -> [f x,  g])

mu     :: (v -> GraphT f v) -> GraphT f v
mu f = Mu (\ ~(x : _) -> [f x])

gfold :: Functor f => (t -> c) -> (([t] -> [c]) -> c) -> (f c -> c) -> Graph f -> c
gfold v l f (Graph g) = trans g where
   trans (Var x) = v x 
   trans (Mu g) = l (map trans . g)
   trans (GIn fa) = f (fmap trans fa)

injectG :: (f :<: g) => f (GraphT g a) -> GraphT g a
injectG  = GIn . inj


cfold :: Functor f => (f t -> t) -> Graph f -> t 
cfold = gfold id (head . fix)

fix :: (a -> a) -> a
fix f = let r = f r in r


------------------------------------------------------
-- Simple pretty printing implementation for graphs --
------------------------------------------------------

newtype StringP = StringP String
instance Show StringP where
    show (StringP s) = s


showGraph :: (Functor f, Show (f StringP)) => [String] -> GraphT f String -> String
showGraph _ (Var v) = v
showGraph ns (Mu f) = show (zipWith showBind names (f names))
    where c = length (f ns)
          (names, ns') = splitAt c ns
          showBind v b = StringP (v ++ " -> " ++ showGraph ns' b)
showGraph ns (GIn t) = "(" ++ show (fmap (StringP . showGraph ns) t) ++ ")"


------------------------------------------------------
-- Better pretty printing implementation for graphs --
------------------------------------------------------


class Functor f => Pretty f where
    pretty :: f String -> f String
    pretty = simpl

simpl :: Functor f => f String -> f String
simpl = fmap semi

semi = ("; " ++) 
paren x = "(" ++ x ++ ")"



prettyGraph' :: (Pretty f, Functor f, Show (f StringP)) => [String] -> GraphT f String -> String
prettyGraph' _ (Var v) = v
prettyGraph' ns (Mu f) = show (zipWith showBind names (f names))
    where c = length (f ns)
          (names, ns') = splitAt c ns
          showBind v b = StringP (v ++ " -> " ++ prettyGraph' ns' b)
prettyGraph' ns (GIn t) = show $ fmap StringP $ pretty $ fmap (prettyGraph' ns) t

prettyGraph (Graph g) = repl $ prettyGraph' (map (\n-> "v" ++ show n) [1..]) g
    where repl (' ' : ';' : r) = ';' : repl r
          repl (x : xs) = x : repl xs
          repl [] = []

instance (Pretty f, Functor f,  Show (f StringP)) =>  Show (Graph f) where
    show g = prettyGraph g
