{-# LANGUAGE TypeOperators, MultiParamTypeClasses, 
             FlexibleInstances, DeriveFunctor, 
             FlexibleContexts, UndecidableInstances,
             GADTs #-}


---------------------
-- Virtual Machine --
---------------------

module VirtualMachine where

import SourceLanguage
import TargetLanguage
import Graph
import Fixpoint
import Prelude hiding (EQ, LT, rem)
import qualified Data.Map as Map


data Elem = VAL Int | VALB Bool | HAN (Conf -> Conf) | HAN'

instance Eq Elem where
    (VAL n) == (VAL m) = n == m
    _ == _ = False

instance Show Elem where
    show (VAL n) = "VAL " ++ show n
    show (VALB n) = "VALB " ++ show n
    show (HAN _) = "HAN"
    show HAN' = "HAN'"

type Stack = [Elem]

type Conf = (Stack, St)

class Exec  f where
    execAlg :: f (Conf -> Conf) -> Conf -> Conf

instance (Exec f, Exec g) => Exec (f :+: g) where
    execAlg (Inl x) = execAlg x
    execAlg (Inr x) = execAlg x


instance Exec ARITH where
    execAlg (PUSH n c) (k, s) = c (VAL n : k, s)
    execAlg (ADD c) (VAL m : VAL n : k, s)    = c (VAL (n + m) : k, s)
    execAlg (ADD _) _ = error "ADD expects two integer elements on the tip of the stack"
    execAlg (MUL c) (VAL m : VAL n : k, s)    = c (VAL (n * m) : k, s)
    execAlg (MUL _) _ = error "MUL expects two integer elements on the tip of the stack"

instance Exec EXCEPT where
    execAlg THROW (k, s) = unwind (k, s)
    execAlg (MARK h c) (k, s) = c (HAN h : k, s)
    execAlg (UNMARK c) (e : HAN _ : k, s) = c (e : k, s)
    execAlg (UNMARK _) _ = error "UNMARK expects a handler element below the tip of the stack"

unwind (HAN h : k, s) = h (k ,s)
unwind (_ : k, s) = unwind (k, s)
unwind ([], s) = ([], s)

instance Exec HALT where
    execAlg HALT cf = cf

instance Exec STATE where
    execAlg (GET v c) (k, s) = case Map.lookup v s of
                                 Just n -> c (VAL n : k, s)
                                 Nothing -> unwind (k, s) 
    execAlg (SET v c) (VAL n : k, s) = c (k, Map.insert v n s)
    execAlg (SET _ _) _ = error "SET expects an integer element on the tip of the stack"

instance Exec COND where
    execAlg (JPC c c') (VALB b : k, s) = if b then c (k, s) else c' (k, s)
    execAlg (JPC _ _) _ = error "JPC expects a Boolean element on the tip of the stack"

instance Exec BOOL where
    execAlg (PUSHB b c) (k, s) = c (VALB b : k, s)
    execAlg (AND c) (VALB m : VALB n : k, s)    = c (VALB (n && m) : k, s)
    execAlg (AND _) _ = error "AND expects two Boolean elements on the tip of the stack"
    execAlg (OR c) (VALB m : VALB n : k, s)    = c (VALB (n || m) : k, s)
    execAlg (OR _) _ = error "OR expects two Boolean elements on the tip of the stack"

instance Exec COMP where
    execAlg (EQ c) (VAL m : VAL n : k, s)    = c (VALB (n == m) : k, s)
    execAlg (EQ _) _ = error "EQ expects two integer elements on the tip of the stack"
    execAlg (LT c) (VAL m : VAL n : k, s)    = c (VALB (n < m) : k, s)
    execAlg (LT _) _ = error "LT expects two integer elements on the tip of the stack"


exec :: (Functor f, Exec f) => Graph f -> Conf -> Conf
exec = cfold execAlg
