\ignore{
\begin{code}
{-# LANGUAGE RankNTypes, StandaloneDeriving, DeriveFunctor, DeriveFoldable,
  DeriveTraversable, TypeOperators, ScopedTypeVariables,
  KindSignatures, MultiParamTypeClasses, GADTs, FlexibleInstances,
  FlexibleContexts, UndecidableInstances, OverlappingInstances,
  NoMonomorphismRestriction
  #-}

import Data.Map (Map, fromList)
import qualified Data.Map as Map
import Data.Traversable
import Data.Foldable (Foldable)
import Control.Monad.State hiding (mapM,join, State)
import Prelude hiding (mapM,rem, LT, EQ)

import Paper hiding (CompAlgG (..), While (..), If (..),
                     Seq(..),Comp(..), Arith(..), State(..))

\end{code}
}


We highlight at this point that all instances of source languages we
have considered thus far have been defined as a single fixpoint of a
signature functor, even though said signature can be compound. We
mention this because at present there are no \textit{restrictions} on
how to combine these different language fragments, permitting all
manner of ill-formed source programs via the abuse of syntax. For
example, we can write:
\begin{spec}
fWhile (fSet (Ref "x") (fVal 10)) (fGet (Ref "x") `fAdd` fVal 1)
\end{spec}

We need to be able to cleanly separate (at least) between
\emph{statements}, which can be combined using sequential composition,
conditionals and loops, and \emph{expressions}, which can be combined
by arithmetic and Boolean operators.

\subsection{Splitting the Source Language}

In order to split the source language into different syntactic
categories -- while retaining its modular properties -- we use
\citeauthor{johann08popl}'s initial algebra semantics of
GADTs~\citep{johann08popl} and combine it with data types \`a la carte
in the style of \citet{bahr11wgp}. The underlying idea is to annotate
each node of the tree type with the syntactic category it resides
in. To this end, each signature functor gets an additional type
argument (and thus ceases to be a functor). For example, using
Haskell's GADT syntax, |Arith| is now defined as follows:
\begin{code}
data Exp  -- type index for expressions

data Arith  e l where   Val  :: Int             -> Arith e Exp
                        Add  :: e Exp -> e Exp  -> Arith e Exp
                        Mul  :: e Exp -> e Exp  -> Arith e Exp
\end{code}

Note that we define an empty datatype |Exp| as a label -- or more
precisely, an index -- for expressions. The |Arith| signature is
simple; each operator only takes expressions and returns
expressions. More interesting is the definition of the signature for
assigning and dereferencing reference cells:
\begin{code}
data Stmt  -- type index for statements

data State e l where    Get  :: Ref           -> State e Exp
                        Set  :: Ref -> e Exp  -> State e Stmt
\end{code}
Note that the |Get| constructor builds an expression, while the |Set|
constructor takes an expression and builds a statement.

As we have already noted, the above indexed signatures are no longer
Haskell functors: instead of mapping types to types, they map functors
to functors (and natural transformations to natural
transformations). In the language of \citet{johann08popl}, these
signatures are \emph{higher-order} functors.

The type constructors |:+:| and |Fix| are easily adjusted to this
setting by also equipping them with an additional type argument:
\begin{code}
data (f ::+ g) (h :: * -> *) e  =  HInl (f h e) |  HInr (g h e)

data HFix f i  = HIn (f (HFix f) i)
\end{code}
\ignore{%
\begin{code}
infixr 8 ::+

instance (HFunctor f, HFunctor g) => HFunctor (f ::+ g) where
    hfmap f (HInl v) = HInl (hfmap f v)
    hfmap f (HInr v) = HInr (hfmap f v)
\end{code}
}%

As expected, the fixpoint of a higher-order functor is itself a type
function of kind |* -> *| (in other words, a family of types). 
In the case of the syntax trees for our target language, this
family comprises the different syntactic categories we want to
represent. Concretely, |HFix (Arith ::+ State) Exp| is the type of
expressions over signature |Arith ::+ State|, whereas |HFix (Arith ::+
State) Stmt| is the corresponding type of statements.

As we have just introduced typed syntax trees, we make use of the
infrastructure in order to also keep track of the object language
typing. Our language is simple enough such that it suffices to add a
type for Boolean expressions:
\begin{code}
data BExp

data Comp e l where    Equ  :: e Exp -> e Exp -> Comp e BExp
                       Lt   :: e Exp -> e Exp -> Comp e BExp
\end{code}


Signatures for control structures are defined in the same style:
\begin{code}
data If    e l where   If     :: e BExp -> e Stmt -> e Stmt ->  If     e Stmt
data While e l where   While  :: e BExp -> e Stmt ->            While  e Stmt

data Seq e l where    Seq :: e Stmt -> e Stmt -> Seq e Stmt
\end{code}
\ignore{%
\begin{code}
instance HFunctor Arith where
      hfmap _ (Val x_afHQ) = Val x_afHQ
      hfmap f_afHW (Add x_afHR x_afHS)
        = Add (f_afHW x_afHR) (f_afHW x_afHS)
      hfmap f_afHX (Mul x_afHT x_afHU)
        = Mul (f_afHX x_afHT) (f_afHX x_afHU)
instance HFunctor While where
      hfmap f_afI9 (While x_afI7 x_afI8)
        = While (f_afI9 x_afI7) (f_afI9 x_afI8)
instance HFunctor If where
      hfmap f_afId (If x_afIa x_afIb x_afIc)
        = If (f_afId x_afIa) (f_afId x_afIb) (f_afId x_afIc)
instance HFunctor Seq where
      hfmap f_afIg (Seq x_afIe x_afIf)
        = Seq (f_afIg x_afIe) (f_afIg x_afIf)
instance HFunctor Comp where
      hfmap f_afIl (Equ x_afIh x_afIi)
        = Equ (f_afIl x_afIh) (f_afIl x_afIi)
      hfmap f_afIm (Lt x_afIj x_afIk)
        = Lt (f_afIm x_afIj) (f_afIm x_afIk)
\end{code}
}%

We now have the infrastructure to define and combine signatures. In
the next section we shall see that the machinery to define folds on
fixpoints and to define smart constructors carries over to the setting
of higher-order functors easily.

\subsection{Folds and Smart Constructors}
\label{sec:folds-smart-constr}

The representation of modular GADTs that we use here is based on
fixpoints of higher-order functors in the style of
\citet{johann08popl}. The key to this representation is that the
higher-order functors, have slightly less structure than what one
would typically associate with higher-order functors. They only
provide a mapping from natural transformations to natural
transformations:
\begin{code}
class HFunctor f where  hfmap :: (g :-> h) -> f g :-> f h
\end{code}
where natural transformations are defined as follows:
\begin{code}
type f :-> g = forall i . f i -> g i
\end{code}
\ignore{%
\begin{code}
infixr 0 :-> -- same precedence as function space operator ->
\end{code}
}%
In the language of \citet{johann08popl}, these higher-order
functors map functors of kind $||*|| \to *$ to functors of kind $||*||
\to *$, which are exactly what is needed to represent GADTs.

Instance declarations for |HFunctor| are, similarly to the typeclass
|Functor|, straightforward. For example, for |State| the instance is
declared as follows:
\begin{code}
instance HFunctor State where
  hfmap f (Get v)    = Get v
  hfmap f (Set v x)  = Set v (f x)
\end{code}

Using this structure, we can define folds on fixpoints of higher-order
functors. Since signatures are indexed (as are their fixpoints), so
are the algebras that are used to define folds. More precisely, given
a higher-order functor |f| and a type constructor |c :: * -> *|, an
|f|-algebra with carrier |c| is a natural transformation of type |f c
:-> c|. Apart from the typing, the implementation of folds is the same
as for ordinary Haskell functors:
\begin{code}
hfold :: HFunctor f => (f c :-> c) -> HFix f :-> c
hfold f (HIn t) = f (hfmap (hfold f) t)
\end{code}

The definition of the typeclass |:<:| is also straightforwardly
lifted to the setting of higher-order functors.
\begin{code}
class (sub :: (* -> *) -> * -> *) ::< sup where
    hinj :: sub a :-> sup a
\end{code}
\ignore{%
\begin{code}
    hprj :: sup a i -> Maybe ((sub a) i)

instance (::<) f f where
    hinj = id
    hprj = Just

instance (::<) f (f ::+ g) where
    hinj = HInl
    hprj (HInl x) = Just x
    hprj (HInr _) = Nothing


instance (f ::< g) => (::<) f (h ::+ g) where
    hinj = HInr . hinj
    hprj (HInr x) = hprj x
    hprj (HInl _) = Nothing
\end{code}
}%

The instance declarations for |:<:| are transcribed one-to-one to the
typeclass  |::<|  without surprises, which in turn gives us the 
corresponding inject function:
\begin{code}
hinject :: (g ::< f) => g (HFix f) :-> HFix f
hinject = HIn `o` hinj
\end{code}

As before, we assume that each constructor of a higher-order signature
functor comes with a corresponding smart constructor, e.g.\:
\begin{code}
hWhile :: (While ::< f) => HFix f BExp  -> HFix f Stmt -> HFix f Stmt
hWhile x y = hinject (While x y)
\end{code}
\ignore{%
\begin{code}
hVal :: (Arith ::< f) => Int -> HFix f Exp
hVal x = hinject (Val x)
hAdd :: (Arith ::< f) => HFix f Exp -> HFix f Exp -> HFix f Exp
hAdd x y = hinject (Add x y)
hMul :: (Arith ::< f) => HFix f Exp -> HFix f Exp -> HFix f Exp
hMul x y = hinject (Mul x y)

hGet :: (State ::< f) => Ref -> HFix f Exp
hGet x = hinject (Get x)
hSet :: (State ::< f) => Ref -> HFix f Exp -> HFix f Stmt
hSet x y = hinject (Set x y)
hIf :: (If ::< f) => HFix f BExp -> HFix f Stmt -> HFix f Stmt -> HFix f Stmt
hIf x y z = hinject (If x y z)
hSeq :: (Seq ::< f) => HFix f Stmt -> HFix f Stmt -> HFix f Stmt
hSeq x y = hinject (Seq x y)

hLt :: (Comp ::< f) => HFix f Exp -> HFix f Exp -> HFix f BExp
hLt x y = hinject (Lt x y)
hEqu :: (Comp ::< f) => HFix f Exp -> HFix f Exp -> HFix f BExp
hEqu x y = hinject (Equ x y)
\end{code}
}%

Given these smart constructors, we can write the factorial program
from section~\ref{sec:graphs} in a typed way as follows:
\begin{code}
fac :: HFix (Arith ::+ State ::+ Seq ::+ While ::+ Comp) Stmt
fac =  hSet y (hVal 1) `hSeq` hWhile (hVal 0 `hLt` hGet x)
              (  hSet y (hGet y `hMul` hGet x) `hSeq` 
                 hSet x (hGet x `hAdd` hVal (-1)) )
   where  x = Ref "x"; y = Ref "y"
\end{code}

\subsection{Compiler}
\label{sec:compiler}

In this section we briefly describe the changes necessary to adapt the
modular compiler implementation from section~\ref{sec:graphs} to this
typed setting.

The key change that we have to deal with is that algebras are now
natural transformations instead of only functions. In essence this
means that algebras have to honour the typing that the higher-order
functors declare. For example, if we were to write an interpreter, the
arguments of |Lt| must evaluate to integers and |Lt| itself then
evaluates to a Boolean.

For the definition of the compiler the typing is not as important,
since the target language is untyped. The typing is only necessary to
ensure the well-formedness of source programs. In order to ignore the
typing information, we introduce a type constructor |K|, which yields
the following definition of the compilation algebra:
\begin{code}
newtype K a i = K {unK :: a}

class (HFunctor f) => CompAlgG_ f g where
  compAlgG_ ::  f (K (GraphT g v -> GraphT g v)) 
                   :-> K (GraphT g v -> GraphT g v)
\end{code}

In order to simplify instance declarations for the compilation
algebra, we remove the |K| type in the result type:
\begin{code}
class (HFunctor f) => CompAlgG f g where
  compAlgG  ::  f (K (GraphT g v -> GraphT g v)) i -> GraphT g v -> GraphT g v
\end{code}
In order to get an algebra of the correct type, we compose |compAlgG| with |K|:
\[
  |K `o` compAlgG :: f (K (GraphT g v -> GraphT g v)) :-> K (GraphT g v -> GraphT g v)|
\]

The final compiler is thus defined as follows:
\begin{code}
compG :: (CompAlgG f g, HALT :<: g) => HFix f i -> Graph g
compG e = MkGraph (unK (hfold (K `o` compAlgG) e) gHALT)
\end{code}
Here we use |unK| to turn the result of the fold, which is of type |K
(GraphT g v -> GraphT g v) i|, into a function of type |GraphT g v ->
GraphT g v|. This function is then applied to a singleton |HALT|
instruction, which serves as the final code continuation.

It thus remains to change the instance declarations for the
compilation algebra such that they fit this new interface.

The lifting to coproducts now uses the type constructor |::+| instead
of |:+:|:
\begin{code}
instance (CompAlgG f g, CompAlgG h g) => CompAlgG (f ::+ h) g where
    compAlgG (HInl x)     = compAlgG x
    compAlgG (HInr y)     = compAlgG y
\end{code}

The changes to the instance declarations for the individual language
fragments are syntactically very simple: we just need to insert the
constructor |K| for each ``recursive'' argument of the language
construct. For example, for the |While| language fragment we define
the following:
\begin{code}  
instance (COND :<: g) => CompAlgG While g where
  compAlgG (While (K b) (K lb)) c = mu (\ v -> b |> gJPC (lb |> Var v) |> c)
\end{code}

\subsection{Modularity}
\label{sec:well-kind-sign}

The use of higher-order functors may seem overkill at first since our
goal was just to separate expressions from statements. Higher-order
functors allow us to encode GADTs, but the relation between
expressions and statements is much simpler: expressions may occur in
statements, but not vice versa. However, we argue -- similarly to
\cite{bahr11wgp} -- that the higher-order functor representation is
much better for modular definitions.

If we were to stick to functors but still wanted to discriminate
between expressions and statements, we would have to make the
expression language a parameter of the statement language, e.g.:
\begin{spec}
data If     exp stmt = If exp stmt stmt
data While  exp stmt = While exp stmt

type Sig = If ((Fix Arith)) :+: While ((Fix Arith))
type Lang = Fix Sig
\end{spec}
This approach complicates composition of signatures, since we have to
make sure that each statement signature receives the same expression
language argument. Moreover, signatures can no longer be extended
compositionally. For example, if we want to extend the signature |Sig|
with comparison operators, we cannot reuse the definition of |Sig|. We
have to define
\begin{spec}
type Sig' = If ((Fix (Arith :+: Comp))) :+: While ((Fix (Arith :+: Compe)))
\end{spec}
Modular algebra definitions become much more difficult as well since
we have to deal with two different levels of modularity, namely on the
expression and the statement level.

The use of higher-order functors makes modular definitions much more
simple and concise as we only have to deal with one level of
modularity. No matter whether we want to extend the statement or the
expression language, we simply use the coproduct |::+| to add the
desired signature. Additionally, we can exploit the open nature of the
kind system. For example, if we want to add floating point
expressions, we simply introduce a new type index |FPExp|:
\begin{code}
data FPExp -- type index for floating point expressions

data Div e l where Div :: e FPExp -> e FPExp -> Div e FPExp

type Lang = HFix (Arith ::+ While ::+ If ::+ State ::+ Div) Stmt
\end{code}

Likewise, also modular algebra definitions remain as simple as in the
untyped case. For example, assuming that we have a suitable
signature |DIV| for the target language, we may extend the compiler
for the new floating point operator by the following declaration:
\begin{code}
instance (DIV :<: g) => CompAlgG Div g where
  compAlgG (Div (K x) (K y))  c = x |> y |> gDIV |> c
\end{code}
\ignore{%
\begin{code}
instance HFunctor Div where
      hfmap f_afHW (Div x_afHR x_afHS)
        = Div (f_afHW x_afHR) (f_afHW x_afHS)

data DIV c = DIV c

gDIV = injectG . DIV    
\end{code}
}%