\documentclass{llncs}
\input{prelude_paper}
%include polycode.fmt
%include format_paper.fmt

\ignore{
\begin{code}
{-# LANGUAGE RankNTypes, StandaloneDeriving, DeriveFunctor, DeriveFoldable,
  DeriveTraversable, TypeOperators, ScopedTypeVariables,
  KindSignatures, MultiParamTypeClasses, GADTs, FlexibleInstances,
  FlexibleContexts, UndecidableInstances, OverlappingInstances,
  NoMonomorphismRestriction
  #-}

module Paper where

import Data.Map (Map, fromList)
import qualified Data.Map as Map
import Data.Traversable
import Data.Foldable (Foldable)
import Control.Monad.State hiding (mapM,join, State)
import Prelude hiding (mapM,rem, LT, EQ)


o = (.) -- use this composition operator in order to type set it as \circ
\end{code}
}


\begin{document}

\title{Pick'n'Fix}
\subtitle{Capturing Control Flow in Modular Compilers}
\titlerunning{Pick'n'Fix}

\author{Laurence~E.~Day\inst{1} \and Patrick~Bahr\inst{2}}
\authorrunning{L.E.~Day, P.~Bahr}
\institute{
  Functional Programming Laboratory,      
  University of Nottingham\\\email{led@@cs.nott.ac.uk}
  \and  
  Department of Computer
  Science, University of Copenhagen\\
  \email{paba@@di.ku.dk}
}

\maketitle

\begin{abstract}
  We present a modular framework for implementing languages with
  effects and control structures such as loops and conditionals. This
  framework enables modular definitions of both syntax and semantics
  as well as modular implementations of compilers and virtual
  machines. In order to compile control structures, in particular
  cyclic ones, we employ Oliveira and Cook's purely functional
  representation of graphs. Moreover, to separate control flow
  features semantically from other language features, we represent
  source languages using Johann and Ghani's encoding of generalised
  algebraic datatypes as fixpoints of higher-order functors. We
  demonstrate the usage of our modular compiler framework with an
  extended running example and highlight the extensibility of our
  modular compiler implementations.
\end{abstract}


\section{Introduction}
\label{sec:introduction}

A compiler is typically decomposed into the various distinct
manipulations and translations of the syntax of a source language,
such as lexing, parsing and code generation. An additional dimension
along which modularity may be exploited is the set of computational
features supported by the source language. 

In previous work~\citep{compilationalacarte,day11tfp}, we combined the
work of \citet{lianghudakjones} on implementing modular interpreters,
and \citeauthor{swierstra08jfp}'s \emph{data types \`{a} la
  carte}~\citep{swierstra08jfp} to create a modular compiler
framework, which we have since built extensively upon. The result to
date is a modular framework allowing the definition and implementation
of new language features without the modification of any existing
definitions. We have demonstrated the implementation of multiple
such features, including exceptions, mutable state and
functional abstraction.

However, this framework currently falls short of properly capturing
the notion of control flow. For example, the compilation module for exception
handling currently duplicates a code continuation in order to 
represent convergent control flow such as seen in conditional statements.
 Moreover, the framework at present cannot sensibly handle cyclic control
 structures such as unbounded loops.

The underlying issue responsible for these shortcomings is the fact
that the modular target language of the compiler is essentially
tree-shaped, and therefore only able to capture tree-shaped control flow.

In this paper, we address this concern by refactoring our modular compilation
 framework such that it operates on target languages that are \emph{graph} 
 structured instead. This modular graph structure allows us to represent 
 sharing and cycles, solving the issue of code duplication and capturing 
 cyclic control structures respectively.

In particular, the contributions of this paper are as follows:

\begin{itemize}
\item We demonstrate the use of \citeauthor{oliveira12icfp}'s
  \emph{structured graphs}~\citep{oliveira12icfp} as the underpinning
  of the modular compiler target language representation.

\item We demonstrate how the additional structure provided by
  structured graphs allows us to properly compile non-cyclic control
  structures in a modular fashion without code duplication.

\item Most importantly, the graph representation allows us to build
  compiler modules for circular control structures, which we
  demonstrate on an imperative language with loops.

\item In order to deal with an imperative language, we use a typed
  representation of the syntax of source languages using
  \citeauthor{johann08popl}'s fixpoint representation of generalised
  algebraic datatypes (GADTs)~\citep{johann08popl}.

\item We demonstrate our modular compiler framework in action by way
  of an extended example.
\end{itemize}

Throughout this paper, we utilise Haskell~\citep{marlow10haskell} as
both a semantic meta-language and an implementation language.  This
paper has been compiled from a literate Haskell source file, and is
available in the associated material from the authors'
webpages.\footnote{\url{http://www.diku.dk/~paba/picknfix.zip}} Due to
the space restrictions we elided the modular semantic definitions of
the source and target languages presented in this paper. Instead,
these semantic definitions can be found in the associated material in
the form of a modular interpreter and a modular virtual machine.


\section{Compilation \`a la Carte}
\label{sec:repr-source-lang}

Our primary goal is to construct a compiler for imperative languages
 with control structures such as loops and conditionals. In this section,
 we illustrate the basic modular compiler framework as presented in previous
work~\citep{compilationalacarte,day11tfp}, before proceeding to identify the
problems this approach presents when dealing with control structures.

The fundamental idea is the representation of a data type as the least 
fixpoint |Fix f| of a functor |f| (also called the \emph{signature} of the data type):
\begin{code}
data Fix f = In (f (Fix f))
\end{code}
The benefit of defining recursive datatypes in this fashion is that
signature functors compose neatly using coproducts:
\begin{code}
data (f :+: g) e = Inl (f e) | Inr (g e)
\end{code}
\ignore{%
\begin{code}
  deriving (Functor)

infixr 8 :+:
\end{code}
}%

As a result, we can define fragments of the source language in
isolation. Below we define language fragments for arithmetic,
comparison operators, sequential composition, state, and conditional
control structures:
\\[-1.5em]
\begin{minipage}[t]{.55\linewidth}
\begin{code}
data Arith  a  =  Val Int  
               |  Add a a   | Mul a a
data Comp   a  =  Equ a a   | Lt a a
data Seq    a  =  Seq a a

newtype Ref = Ref String   -- reference cells
\end{code}  
\end{minipage}
\begin{minipage}[t]{.4\linewidth}
\begin{code}
data State  a  =  Set Ref a 
               |  Get Ref 
data If     a  =  If a a a
data While  a  =  While a a 
\end{code}    
\end{minipage}
\ignore{%
\begin{code}
deriving instance Functor Arith
deriving instance Functor Comp
deriving instance Functor Seq
deriving instance Functor State
deriving instance Functor If
deriving instance Functor While
\end{code}
}%

The source language supporting all of the above featues is thus 
represented by the type |Fix (Arith :+: Comp :+: Seq :+: State If :+: While)|.

In order to define functions over datatypes constructed in such a
modular manner, we make use of a generic fold operator:
\begin{code}
fold :: Functor f => (f c -> c) -> Fix f -> c 
fold f (In t) = f (fmap (fold f) t)
\end{code}
The first argument to the above is called an \emph{|f|-algebra}, and
is intuitively viewed as a directive for recursively processing data
constructors. The type |c| of an |f|-algebra is called its
\emph{carrier} and forms the result type of a fold.

To circumvent the need to manually tag values with |Inl| and |Inr|
constructors when introducing elements of a compound functor
constructed using |:+:|, the data type \`a la carte technique makes
use of a typeclass |:<:|, which provides an automatic injection
mechanism:

\noindent
\begin{minipage}{.45\linewidth}
\begin{code}
class g :<: f where
  inj :: g a -> f a
\end{code}  
\end{minipage}
\begin{minipage}{.45\linewidth}
\begin{code}
inject :: (g :<: f) => g (Fix f) -> Fix f
inject = In `o` inj
\end{code}
\end{minipage}

\ignore{%
\begin{code}
instance (:<:) f f where
  inj = id
  
instance (:<:) f (f :+: g) where
  inj = Inl
  
instance ((:<:) f h) => (:<:) f (g :+: h) where
  inj = Inr . inj
\end{code}
}%
We do not define the instance declarations for |:<:| here, but instead
refer the interested reader to the original
work~\citep{swierstra08jfp}. Suffice it to say that we have |g :<: f|
if and only if |f| is a sum with |g| a summand in |f|.

To further simplify the construction of elements of a
fixpoint |Fix f|, we make use of |smart constructors| of the following form:
\begin{code}
fIf :: (If :<: f) => Fix f -> Fix f -> Fix f -> Fix f  
fIf c i t = inject (If c i t)
\end{code}
\ignore{%
\begin{code}
fVal :: (Arith :<: f) => Int -> Fix f
fVal x = inject (Val x)
fAdd :: (Arith :<: f) => Fix f -> Fix f -> Fix f
fAdd x y = inject (Add x y)
fMul :: (Arith :<: f) => Fix f -> Fix f -> Fix f
fMul x y = inject (Mul x y)
fWhile :: (While :<: f) => Fix f -> Fix f -> Fix f
fWhile c i = inject (While c i)

fGet :: (State :<: f) => Ref -> Fix f
fGet x = inject (Get x)
fSet :: (State :<: f) => Ref -> Fix f -> Fix f
fSet x y = inject (Set x y)
fSeq :: (Seq :<: f) => Fix f -> Fix f -> Fix f
fSeq x y = inject (Seq x y)
fLt :: (Comp :<: f) => Fix f -> Fix f -> Fix f
fLt x y = inject (Lt x y)
fEqu :: (Comp :<: f) => Fix f -> Fix f -> Fix f
fEqu x y = inject (Equ x y)

fPUSH  n c = inject (PUSH n c)
fADD    c = inject (ADD c)
fMUL    c = inject (MUL c)
fJPC  = \ l c -> inject (JPC l c)
fGET v c = inject (GET v c)
fSET v c = inject (SET v c) 
fHALT = inject HALT
\end{code}
}%

For example, we can write the following program in our extensible language:
\begin{spec}
let x = Ref "x"
in fWhile (fGet x `fLt` fVal 10) (fSet x (fGet x `fAdd` fVal 1)) 
\end{spec}
The program increments the reference cell |x| until it reaches the
value |10|.

Next, we show how to use this framework in order to implement a compiler in
a modular manner. To this end, we first define the appropriate signatures for
the |target| language. For now we only consider the arithmetic and conditional
language signatures:
\begin{code}
data ARITH  e = PUSH Int e | ADD e | MUL e
data COND   e = JPC e e
data HALT   e = HALT
\end{code}

Intuitively, a |PUSH| instruction places its associated integer on top
of the current stack, and the |ADD| and |MUL| replace the two topmost
elements of the stack by their sum respectively their product.  The
|JPC| instruction of the conditional signature pops the first integer
off the stack and executes its first argument if this integer is
non-zero; otherwise it executes the second argument. Finally, the
|HALT| instruction simply ends the computation.

Secondly, we identify the structure of the compilation algebra. In
this paper, we define our compilers using \textit{continuation-passing
  style} (CPS), requiring the presence of an accumulator argument:
\begin{code}
class (Functor f) => CompAlg f g where
    compAlg :: f (Fix g -> Fix g) -> Fix g -> Fix g
\end{code}
The final compiler is obtained by folding over the compilation algebra
defined above and providing the |HALT| instruction as an initial
continuation.
\begin{code}
comp :: (HALT :<: g, CompAlg f g) => Fix f -> Fix g
comp p = fold compAlg p fHALT
\end{code}


In order to simplify the presentation of our CPS compiler, we define
an explicit function application operator (similar to Haskell's
standard |$| operator): %$
\begin{code}
infixr 0  |>

(|>) :: (a -> b) -> a -> b
f |> x = f x
\end{code}

Finally, we have to declare instances of the compilation algebra. The
following declaration derives instances for compound source language
signatures:
\begin{code}
instance (CompAlg f h, CompAlg g h) => CompAlg (f :+: g) h where
    compAlg (Inl  x  )  = compAlg x
    compAlg (Inr  y  )  = compAlg y
\end{code}
Then we only need to declare instances for each of the individual
source language signatures. We start with arithmetic and conditionals:
\begin{code}
instance (ARITH :<: g) => CompAlg Arith g where
    compAlg (Val n)    c = fPUSH n |> c
    compAlg (Add x y)  c = x |> y |> fADD |> c
    compAlg (Mul x y)  c = x |> y |> fMUL |> c

instance (COND :<: g) => CompAlg If g where
    compAlg (If b x y)    c = b |> fJPC (x |> c) (y |> c)
\end{code}

We can thus instantiate the modular compiler for the implemented
language features, e.g.:
\begin{code}
compArithIf :: Fix (Arith :+: If) -> Fix (ARITH :+: COND :+: HALT)
compArithIf = comp
\end{code}

There is a subtle issue in the instantiation for |If|, however: the
target code produced for |If| first evaluates the Boolean
condition before executing the corresponding branch. However, after
that, control flow returns to the continuation |c|. Due to
the limitations of the target language, this continuation must be duplicated.
 A similar duplication occurs in the compilation of exception handlers as seen
 in previous work~\citep{compilationalacarte,day11tfp}

A more pressing issue becomes apparent if we wish to define an
instance of the compilation algebra for |While|:
\begin{spec}
instance (COND :<: g) => CompAlg While g where
    compAlg (While b lb)  c = b |> fJPC (lb |> ...) c
\end{spec}
What we require in place of the ellipsis is an instruction which
jumps back to the check of the condition |b|, after executing the loop body |lb|.
 The tree structure of the target language does not allow us to represent
 this cyclic control flow.

The core idea we present here in order to fix the issues described above
 is simple: endow the target language with a graph structure. In the next 
 section, we illustrate how to do this whilst retaining the modular properties
of the target language and the compiler implementation.

\section{A Graph-Structured Target Language}
\label{sec:graphs}

Typically, a graph-structured target language for a compiler uses
explicit jumps and labels (as seen in, for example, Hoopl~\citep{ramsey10haskell}).
 This style of representing graphs, whilst close to physical machine 
 implementations, is quite tedious to use and error prone when humans become involved.
 Instead, we shall use the purely functional representation of graphs by 
 \citet{oliveira12icfp}, dubbed \emph{structured graphs}. In short, structured 
 graphs provide a representation of term graphs, which uses an elegant encoding of
sharing and cyclicity using \emph{parametric higher-order abstract
  syntax}~\citep{chlipala08icfp}. This representation provides a
simple interface to construct graphs in a compositional fashion (at
the cost of a more complicated and restrictive interface for
`consuming' them).
\subsection{From Fixpoints to Graphs}
\label{sec:extendingfixpoints}

Structured graphs, at their core, represent \emph{term graphs} using
mutually recursive let bindings. To this end the definition of
structured graphs extends the definition of fixpoints by two
additional constructors, |Var| and |Mu|, for metavariables and mutually
recursive bindings, respectively:
\begin{code}
data GraphT f v  =  Var v
                 |  Mu ([v] -> [GraphT f v])
                 |  GIn (f (GraphT f v))
\end{code}
The additional type variable |v| is used as the type for the
metavariables in the graph. |Mu| represents binders using higher-order
abstract syntax (HOAS). In order to allow mutually recursive bindings,
|Mu| is a function that takes a list of metavariables and returns a
list of corresponding graphs. The intention is that the $i$-th
metavariable in the argument list is bound to the $i$-th graph in the
result list. The entry point is the first graph in the result list.

In this paper, we only use the above |Mu| constructor in two distinct
 cases, viz.\ non-recursive let bindings and fixpoints over a
single argument, which we express as follows:
\begin{code}
letx :: GraphT f v -> (v -> GraphT f v) -> GraphT f v
letx g f = Mu (\ ~(_ : x : _) -> [f x,  g])

mu :: (v -> GraphT f v) -> GraphT f v
mu f = Mu (\ ~(x : _) -> [f x])
\end{code}

With these two combinators we may represent a non-recursive let
binding ``|let x = b in s|'' as |letx b (\ x -> s)|, and a fixpoint of
|f| as |mu f|.

Structured graphs use a restricted form of HOAS called
\emph{parametric HOAS}~\citep{chlipala08icfp}. When constructing
structured graphs the type |v| of metavariables is left
polymorphic. To ensure this, structured graphs are wrapped in the
following type, which enforces the parametric polymorphism of the
metavariables type |v|:
\begin{code}
newtype Graph f = MkGraph (forall v . GraphT f v)
\end{code}
The parametricity of the type of metavariables ensures that the
function argument passed to the |Mu| constructor is indeed only used
for defining binders.

In analogy to smart constructors for the fixpoint type constructor
|Fix|, we use smart constructors to simplify the construction of
graphs. To this end we transcribe |inject| from fixpoints to graphs:
\begin{code}
injectG :: (f :<: g) => f (GraphT g a) -> GraphT g a
injectG  = GIn `o` inj
\end{code}
Using |injectG|, we obtain smart constructors such as the following:
\begin{code}
gPUSH :: (ARITH :<: f) => Int -> GraphT f v -> GraphT f v
gPUSH n c = injectG (PUSH n c)
\end{code}
\ignore{%
\begin{code}
gADD    c = injectG (ADD c)
gMUL    c = injectG (MUL c)
gEQ    c = injectG (EQ c)
gLT    c = injectG (LT c)
gJPC  = \ l c -> injectG (JPC l c)
gGET v c = injectG (GET v c)
gSET v c = injectG (SET v c) 
gHALT = injectG HALT
\end{code}
}%

% Note that the structured graphs that we have defined above use
% ordinary functors as opposed to higher-order functors. Building graph
% types based on higher-order functors is possible too as
% \citet{oliveira13pepm} showed. However, for the purpose we have in
% mind, viz.\ representing control-flow graphs, this is not
% necessary. Nevertheless, one could imagine a use of a more richly
% typed target language, whose types encode invariants about the state
% of the stack (see for example \citet{mckinna06jfp}), but this is an
% issue orthogonal to the goals of this paper.

\subsection{Compiling To Graphs}
\label{sec:compilinggraphs}

We can now refactor the carrier of the compilation algebra to reflect
the usage of structured graphs instead of least fixpoints:
\begin{code}
class Functor f => CompAlgG f g where
  compAlgG :: f (GraphT g v -> GraphT g v) -> GraphT g v -> GraphT g v
\end{code}

As before the type class is easily lifted to coproducts:
\begin{code}
instance (CompAlgG f g, CompAlgG h g) => CompAlgG (f :+: h) g where
    compAlgG (Inl x)     = compAlgG x
    compAlgG (Inr y)     = compAlgG y
\end{code}

Note that we use the |GraphT| type constructor instead of |Graph|. As
a general rule of thumb, we use |GraphT| in order to build up graphs,
and after that we use the |MkGraph| constructor to construct a graph of
type |Graph g|. The rank 2 polymorphic type of |MkGraph| ensures that
the construction of the underlying graph of type |GraphT| was indeed
polymorphic in the type |v| of metavariables:
\begin{code}
compG :: (CompAlgG f g, HALT :<: g) => Fix f -> Graph g
compG e = MkGraph (fold compAlgG e gHALT)
\end{code}

The implementation of the compiler for |ARITH| is analogous to the
fixpoint version from section~\ref{sec:repr-source-lang}. The only
aspect that has to be changed is that we use the smart constructors
for graphs:
\begin{code}
instance (ARITH :<: g) => CompAlgG Arith g where
  compAlgG (Val n)    c = gPUSH n |> c
  compAlgG (Add x y)  c = x |> y |> gADD |> c
  compAlgG (Mul x y)  c = x |> y |> gMUL |> c
\end{code}

In principle, we may transcribe the compiler definition for |If|
in the same simple way. However, we can exploit the sharing
capabilities that the graph structure affords us:
\begin{code}
instance (COND :<: g) => CompAlgG If g where
  compAlgG (If b p1 p2) c = letx c 
    (\ v ->  b |> gJPC (p1 |>  Var v) |> p2 |> Var v)
\end{code}

Instead of placing the code continuation |c| directly into the
generated code (and thereby duplicating |c|), we bind |c| to the
metavariable |v|, which is then used instead of |c|, thus avoiding the
duplication.

Likewise, for compiling loops, we make use of the graph structure,
too. This time, however, we need to construct a cycle, for which we
use the |mu| combinator:
\begin{code}  
instance (COND :<: g) => CompAlgG While g where
  compAlgG (While b lb) c = mu (\ v -> b |> gJPC (lb |> Var v) |> c)
\end{code}

Next, we give the definition for compiling the signatures |State| and
|Comp|. To do this we need corresponding instructions in the target
language:
\begin{code}
data STATE  e  = GET Ref e | SET Ref e
data COMP   e  = EQ e | LT e
\end{code}
\ignore{%
\begin{code}
deriving instance Functor STATE
deriving instance Functor COMP
\end{code}
}%
The corresponding compiler definitions are straightforward:
\begin{code}
instance (STATE :<: g) => CompAlgG State g where
  compAlgG (Get v)    c  = gGET v |> c
  compAlgG (Set v e)  c  = e |> gSET v |> c

instance (COMP :<: g) => CompAlgG Comp g where
    compAlgG (Equ x y)  c = x |> y |> gEQ |> c
    compAlgG (Lt x y)   c = x |> y |> gLT |> c
\end{code}


Finally, we give the definition for compiling sequential composition:
\begin{code}
instance CompAlgG Seq g where
  compAlgG (Seq x y) c = x |> y |> c
\end{code}

To see the resulting compiler in action, we consider the following example
source program that computes the factorial:
\begin{code}
type Source = Arith :+: State :+: Seq :+: While :+: Comp

fac :: Fix Source
fac =  fSet y (fVal 1) `fSeq` fWhile (fVal 0 `fLt` fGet x)
              (  fSet y (fGet y `fMul` fGet x) `fSeq` 
                 fSet x (fGet x `fAdd` fVal (-1)) )
   where  x = Ref "x"; y = Ref "y"
\end{code} 
 
In order to determine the target language, we need only sum up the
constraints on the target signatures in the declarations of |CompAlgG|
for the individual source language features:
\begin{code}
type Target = ARITH :+: STATE :+: COND :+: COMP :+: HALT

compSource  ::  Fix Source ->  Graph Target
compSource = compG
\end{code}

Note that Haskell's type system makes sure that the target language
has the instructions required to compile the source language. For
instance, if we forget to include |COMP| in the target language,
 Haskell's type checker complains:
\begin{verbatim}
No instance for (COMP :<: HALT) arising from a use of `comp'
\end{verbatim}

Applying |compSource| to |fac| returns the following graph: 
\begin{verbatim}
 PUSH 1; SET y; [v1 -> PUSH 0; GET x; LT; JPC (GET y; GET x; MUL; 
   SET y; GET x; PUSH (-1); ADD; SET x; v1); HALT]
\end{verbatim}
The code inside the brackets corresponds to bindings in the graph
structure constructed by |Mu|: the metavariable \texttt{v1} is bound
to the code to the right of the \texttt{->} arrow, which in turn
contains a reference to \texttt{v1}. Thus, as expected, the output
code graph has a single cycle, corresponding to the loop of the source
program.


\section{Typing the Source Langage}
\label{sec:typing-source-lang}

%include typed.lhs


\section{Discussion}
\label{sec:discussion}


\subsection{Future Research Avenues}

\paragraph{Additional Computational Constructs} We would like to
describe other computational features in the manner that we have
presented in order to further increase the expressive power of a
modular source language.  In particular, we are interested in
implementing I/O, explicit parallelism and continuations in this
manner, and exploring how their semantics interacts with the existing
framework.

\paragraph{Data-Flow Analysis and Optimisations} A natural extension
of our work is to implement data-flow analysis and optimisations in a
modular style as well. A good starting point for extending our work in
this direction is \emph{Hoopl}, the \emph{Higher Order Optimisation
  Library} of \citet{ramsey10haskell}. Hoopl is a Haskell library that
allows compiler implementers to define data-flow analyses and
implement optimising transformations that are informed by this
analysis. Modular implementations of optimising transformations can be
achieved using the same techniques as presented in this paper. In
particular, data-flow analysis and the underlying lattice structures
can be defined in a modular way for (at least) standard textbook
analyses.

\paragraph{Testing and Reasoning} 
An important property of a compiler is its trustworthiness. Does it
perform only semantics preserving transformations? Establishing such
trustworthiness in a modular fashion as well still remains a
considerable challenge. However, using the same techniques as
presented here, automatic test case generation (i.e.\ generation of
input programs and initial configurations) can be implemented in a
modular fashion.  Rigorous and machine-checked correctness proofs,
however, require new reasoning techniques that work in a modular
setting. There is a growing interest in formalising programming
language metatheory in a modular
fashion~\citep{delaware13popl,delaware13icfp,schwaab13plpv}. However,
building modular proofs of compiler correctness have to deal with
additional difficulties. Such proofs have to be modular along both the
source and target language as well as computational effects. As the
work of \citet{delaware13icfp} shows, modular reasoning about effects
already becomes a considerable obstacle for type soundness proofs.

\subsection{Related Work}
\label{sec:related-work}

\paragraph{Modular Compilers}
\label{sec:modular-compilers}

William Harrison's doctoral thesis~\citep{harrisonthesis} focuses on
the topic of modular compilation by defining `reusable compiler
building blocks'. These blocks are presented using two distinct
approaches, one of which is \emph{monadic code generators}, which are
closely related to our own research. Also contained within each block
is a set of equations defining the `compilation semantics' of the
feature in question, and the compiler is verified correct by
formulating and proving relations between the standard and compilation
semantics.

Marcos Viera's doctoral thesis~\citep{firstclass} presents the usage of
attribute grammars as the vehicle for defining extensible programming
languages. In short, individual fragments of an attribute grammar can
describe individual features of a language, and compositionality is
guaranteed via the Haskell type system. There is also a number of
standalone attribute grammar systems with particular focus on
extensible language implementation such as LISA~\citep{mernik05clss},
JastAdd~\citep{ekman07scp} and Silver~\citep{vanwyk10scp}.

\paragraph{Graph Representations for Compilers}
\label{sec:graph-repr-comp}

As previously mentioned, the data-flow analysis library
Hoopl~\citep{ramsey10haskell} uses a graph representation based on
explicit labels and jumps. More recently, \citet{bahr14flops}
demonstrated the use of structured graphs for representing the target
language of a compiler. However, \citeauthor{bahr14flops}'s work is
focused on compiler verification and does not consider cyclic
graphs. Moreover, neither \citet{ramsey10haskell} nor
\citet{bahr14flops} consider the problem of modularising the target
language.

\paragraph{Recursion Schemes}
\label{sec:recursion-schemes}
The recursion schemes used to define the compiler and the semantics
heavily influences how compiler and semantics can be extended. More
structured recursion schemes derived from tree
automata~\citep{bahr12mpc,bahr13wgp} and attribute
grammars~\citep{viera09icfp,viera12sblp}
offer more freedom to \emph{replace} parts of modular definition as
opposed to only being able to \emph{extend} them. In the same
direction goes the work of \citet{kimmell05ecbs} and
\citet{frisby10scp} who introduce algebra combinators such as switch
and sequence algebras to compose algebras.

\paragraph{Modular Syntax} The use of higher-order functors to
represent indexed datatypes and families of mutually recursive
datatypes stems from \citet{johann08popl}. \citet{yakushev09icfp}
applied this technique to generic
programming. \citet{bahr11wgp,bahr12msfp} employed fixpoints of
higher-order functors to represent modular syntax and semantics, and
combined it with parametric HOAS. \citet{axelsson12icfp} introduces a
different approach to modular well-typed definitions of syntax and
semantics: he developed an applicative encoding of the syntax that
makes heavy use of type indexing to describe the signature of
individual language features. \citet{oliveira13pepm} used structured
graphs to represent source languages. To this end they combined
structured graphs with the typing mechanism of indexed types and
modular definitions using type classes.



% We recommend abbrvnat bibliography style.

\bibliographystyle{splncsnat}
% The bibliography should be embedded for final submission.

\bibliography{../common/main}


\end{document}

%%% Local Variables: 
%%% mode: latex
%%% TeX-master: t
%%% End: 
